// (C) Copyright Cryolite 2005.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#if !defined(CRY_DELETER_DEALLOCATOR_HPP_INCLUDE_GUARD)
#define CRY_DELETER_DEALLOCATOR_HPP_INCLUDE_GUARD

#include <boost/compressed_pair.hpp>


namespace cry{

template<class Allocator>
class deallocator
{
  // All the member functions never throw.
  // CONSIDER: This guarantee relies on the assumption that `Allocator' never
  //           throw in its copy constructor. Is it reasonable?
public:
  typedef Allocator allocator_type;
  typedef typename allocator_type::value_type value_type;
  typedef typename allocator_type::reference reference;
  typedef typename allocator_type::const_reference const_reference;
  typedef typename allocator_type::pointer pointer;
  typedef typename allocator_type::const_pointer const_pointer;
  typedef typename allocator_type::size_type size_type;
  typedef typename allocator_type::difference_type difference_type;
  
  deallocator(size_type n, allocator_type const &a = allocator_type())
    : size_and_allocator_(n, a)
  {}

  deallocator(deallocator const &r)
    : size_and_allocator_(r.size_and_allocator_)
  {}

  size_type size() const
  {
    return size_reference();
  }

  allocator_type get_allocator() const
  {
    return allocator_reference();
  }
  
  // Are there any cases where const_pointer should be passed to this?
  void operator()(pointer p) const
  {
    if(size_reference() != 0){
      // allocator_refrence() cannot be used, because deallocate is a non-const
      // member function.
      get_allocator().deallocate(p, size_reference());
    }
  }

private:
  size_type &size_reference()
  {
    return size_and_allocator_.first();
  }

  size_type const &size_reference() const
  {
    return size_and_allocator_.first();
  }
  
  allocator_type &allocator_reference()
  {
    return size_and_allocator_.second();
  }

  allocator_type const &allocator_reference() const
  {
    return size_and_allocator_.second();
  }

private:
  boost::compressed_pair<size_type, allocator_type> size_and_allocator_;
}; // class deallocator

} // namespace cry


#endif // !defined(CRY_DELETER_DEALLOCATOR_HPP_INCLUDE_GUARD)
