// (C) Copyright Cryolite 2005.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#if !defined(CRY_MOVE_IS_MOVE_CONVERTIBLE_HPP_INCLUDE_GUARD)
#define CRY_MOVE_IS_MOVE_CONVERTIBLE_HPP_INCLUDE_GUARD

#include <boost/mpl/and.hpp>
#include <boost/mpl/not.hpp>

#include <boost/type_traits/is_same.hpp>
#include <boost/type_traits/remove_reference.hpp>

#include <boost/move.hpp>

#include <cry/type_traits/is_const.hpp>
#include <cry/type_traits/plain.hpp>


// NOTE: This trait should be specialized for a type which has a move
//       conversion.
namespace cry{

template<class T, class U> // T and U may be (reference to) cv-qualified type.
struct is_move_convertible
  : boost::mpl::and_<
      boost::mpl::not_<
        cry::is_const<typename boost::remove_reference<U>::type>
      >
    , boost::is_same<
        typename cry::plain<T>::type, typename cry::plain<U>::type
      >
    , boost::is_movable<T>
    >
{};

} // namespace cry


#endif // !defined(CRY_MOVE_IS_MOVE_CONVERTIBLE_HPP_INCLUDE_GUARD)
