// (C) Howard Hinnant 2004
// (C) Jonathan Turkanis 2004
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt.)

//
// Contains type traits machinery for incomplete arrays. MPL compatibility
// is included for completeness, but is not necessary for the current 
// application.
// 

// Note: This implementation is borrowed from
// http://www.kangaroologic.com/move_ptr/libs/move_ptr/doc/index.html
//
//                                                    2005 Cryolite

#ifndef CRY_MOVE_PTR_IS_CONVERTIBLE_HPP_INCLUDE_GUARD
#define CRY_MOVE_PTR_IS_CONVERTIBLE_HPP_INCLUDE_GUARD

#include <boost/config.hpp> // BOOST_STATIC_CONSTANT.

#include <boost/mpl/and.hpp>
#include <boost/mpl/eval_if.hpp>

#include <boost/type_traits/is_array.hpp>
#include <boost/type_traits/is_same.hpp>
#include <boost/type_traits/is_convertible.hpp>
#include <boost/type_traits/remove_cv.hpp>
#include <boost/type_traits/remove_bounds.hpp>

#include <boost/utility/enable_if.hpp>


namespace cry { namespace move_ptrs {

// From Howard Hinnant.
template<typename T, typename U>
struct is_array_convertible{
  typedef typename boost::remove_bounds<T>::type      t_element;
  typedef typename boost::remove_bounds<U>::type      u_element;
  typedef typename boost::remove_cv<t_element>::type  t_base;
  typedef typename boost::remove_cv<u_element>::type  u_base;
  typedef typename boost::mpl::and_<
      boost::is_array<T>
    , boost::is_array<U>
    , boost::is_same<t_base, u_base>                  // CONSIDER: ???
    , boost::is_convertible<t_element *, u_element *> // CONSIDER: ???
    >::type type;
  BOOST_STATIC_CONSTANT(bool, value = type::value);
};

template<typename T, typename U>
struct is_convertible
  : boost::mpl::eval_if<
      boost::is_array<T>
    , is_array_convertible<T, U>
    , boost::is_convertible<T *, U *> 
    >::type
{ };

template<typename Src, typename Tgt, typename T = void>
struct enable_if_convertible 
  : boost::enable_if< 
      is_convertible<Src, Tgt>
    , T
    >
{ };

}} // namespace cry::move_ptrs


#endif // !defined(CRY_MOVE_PTR_IS_CONVERTIBLE_HPP_INCLUDE_GUARD)
