// (C) Copyright Cryolite 2004.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
//  The meta-function defined in this header differs from Boost.Range's one.
// First, the constness of argument affects the result. In this point of view,
// this meta-function is equivalent to `boost::range_result_iterator.' See also
// 
// http://thread.gmane.org/gmane.comp.lib.boost.devel/123294
//
// Second, this meta-function has an empty definition as its default. This
// means that it can be used in SFINAE contexts. IOW, even if an invalid type is
// substituted to its argument, it does not cause a compile error and just
// eliminates overloads which cause invalid substitutions from overload
// resolution set.

#if !defined(CRY_RANGE_ITERATOR_HPP_INCLUDE_GUARD)
#define CRY_RANGE_ITERATOR_HPP_INCLUDE_GUARD

#include <cstddef>
#include <utility>

#include <boost/type_traits/is_const.hpp>

#include <cry/introspection/has_iterator_mt.hpp>
#include <cry/introspection/has_const_iterator_mt.hpp>


namespace cry{ namespace range{

namespace detail{

template<class T, bool has_iterator>
struct mutable_iterator_impl
{ };

template<class Iterator>
struct mutable_iterator_impl<std::pair<Iterator, Iterator>, false>
{
  typedef Iterator type;
};

template<class T, std::size_t sz>
struct mutable_iterator_impl<T[sz], false>
{
  typedef T *type;
};

template<>
struct mutable_iterator_impl<char *, false>
{
  typedef char *type;
};

template<>
struct mutable_iterator_impl<char const *, false>
{
  typedef char const *type;
};

template<>
struct mutable_iterator_impl<wchar_t *, false>
{
  typedef wchar_t *type;
};

template<>
struct mutable_iterator_impl<wchar_t const *, false>
{
  typedef wchar_t const *type;
};

template<class C>
struct mutable_iterator_impl<C, true>
{
  typedef typename C::iterator type;
};

template<class T, bool has_const_iterator>
struct immutable_iterator_impl
{ };

template<class Iterator>
struct immutable_iterator_impl<std::pair<Iterator, Iterator> const, false>
{
  typedef Iterator type;
};

template<class T, std::size_t sz>
struct immutable_iterator_impl<T const [sz], false>
{
  typedef T const *type;
};

template<class C>
struct immutable_iterator_impl<C, true>
{
  typedef typename C::const_iterator type;
};

template<class T, bool is_const>
struct iterator_impl0
  : public mutable_iterator_impl<
      T
    , has_iterator_mt<T>::value
    >
{ };

template<class T>
struct iterator_impl0<T, true>
  : public immutable_iterator_impl<
      T
    , has_const_iterator_mt<T>::value
    >
{ };

} // namespace detail

template<class T>
struct iterator
  : public detail::iterator_impl0<
      T
    , boost::is_const<T>::value
    >
{ };

template<class T, int sz>
struct iterator<T [sz]>
{
  typedef T *type;
};

template<class T, int sz>
struct iterator<T const [sz]>
{
  typedef T const *type;
};

}} // namespace cry::range


#endif // !defined(CRY_RANGE_ITERATOR_HPP_INCLUDE_GUARD)
